#include "ipEnhanceDCT2D.h"

namespace Torch {

ipEnhanceDCT2D::ipEnhanceDCT2D(int block_size_, const char *coding_, int dct_dim_, real lambda_) : ipDCT2D(block_size_, coding_, dct_dim_)
{
	lambda = lambda_;
	n_band = 2 * block_size - 1;
	d = (real *)allocator->alloc(block_size * block_size * sizeof(real));
	E = (real *)allocator->alloc(n_band * sizeof(real));
	H = (real *)allocator->alloc(n_band * sizeof(real));
}

void ipEnhanceDCT2D::process(Sequence *seq_in)
{
	real *dst = NULL;

	if(dct_dim > block_size * block_size)
		error("ipDCT2D(): dct_dim (%d) exceeds the total number of coefficients.", dct_dim);

	bool verbose = false;

	//
	// Computes all the coefficients
	if(verbose) print("Computes all the coefficients ...\n");
	dst = d;
	for(int i = 0; i < block_size * block_size ; i++) 
	{
	   	int u = dct_coeff_order[2*i];
		int v = dct_coeff_order[2*i+1];
	
		real sum = 0.0;
		
		for(int y = 0; y < block_size; y++)  
		{
		   	int offset_ = y * block_size;
			
		   	for(int x = 0; x < block_size; x++)  
		   	{
			   	int idx_ = offset_ + x;
				
				sum += seq_in->frames[0][idx_] * cosine[u][v][x][y];
			}
		}
	
		*dst++ = alpha[u][v] * sum;
	}
	
	
	//
	// Computes bands
	if(verbose) print("Computes %d bands ...\n", n_band);
	for(int n = 0 ; n < n_band ; n++) E[n] = 0.0;
	
	for(int v = 0; v < block_size; v++)
	{
	   	int offset_ = v * block_size;
		
		for(int u = 0; u < block_size; u++)
		{
		   	int t = u+v;
			
			int idx_ = offset_ + u;
			
			E[t] += fabs(d[idx_]);
		}
	}

	if(verbose) for(int n = 0 ; n < n_band ; n++) print(" n=%d \t E_n=%g\n", n, E[n]);
	
	//
	// Normalize bands
	if(verbose) print("Normalizes bands ...\n");
	for(int n = 0 ; n < n_band ; n++) H[n] = 0.0;

	real normEsum = E[0];
	real Esum = E[0];
	//H[0] = E[0];
	for(int n = 1 ; n < n_band ; n++)
	{
		H[n] = normEsum / Esum;
		
		real z = lambda * H[n] * E[n];
	   	normEsum += z;
	   	Esum += E[n];
		
		if(verbose) print(" n=%d \t E_n=%g \t norm E_n=%g \t H_n=%g\n", n, E[n], z, H[n]);
	}

	
	//
	// Normalize the coefficients
	if(verbose) print("Normalize coefficients ...\n");
	dst = seq_out->frames[0];
	real *d_ = d;
	int iszero = (int) *d_;

	if(iszero == 0)
	{
		for(int i = 0; i < dct_dim ; i++) *dst++ = 0.0;
		
		warning("ipEnhanceDCT2D::process() first coefficient is zero -> setting other coefficients to zero");
	}
	else
	{ 
		*dst++ = *d_++;

		for(int i = 1; i < dct_dim ; i++) 
		{
	   		int u = dct_coeff_order[2*i];
			int v = dct_coeff_order[2*i+1];

			*dst++ = lambda * (*d_++) * H[u+v];
		}
	}
}

ipEnhanceDCT2D::~ipEnhanceDCT2D()
{
}

}

