#ifndef IPROTATE_INC
#define IPROTATE_INC

#include "vision.h"
#include "Image.h"
#include "ipCore.h"
#include "ipShift.h"

namespace Torch {

/** This class is designed to rotate an image

    It implements the rotation by shear also called Paeth rotation.

    \begin{verbatim}
	@inproceedings{paeth:86,
	author = {A W Paeth},
	title = {A fast algorithm for general raster rotation},
	booktitle = {Proceedings on Graphics Interface '86/Vision Interface '86},
	year = {1986},
	pages = {77--81},
	location = {Vancouver, British Columbia, Canada},
	publisher = {Canadian Information Processing Society},
	}
    \end{verbatim}
  
    The original source code can be found \URL[here]{http://www.codeproject.com/bitmap/rotatebyshear.asp}

    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class ipRotate : public ipCore
{
   	// width of the input image
	int width;

	// height of the input image
	int height;

	// width of the output image
	int width_out;

	// height of the output image
	int height_out;

	// computes internal variables
	void init_(void);

	//
	struct ipRotateMatrix2D
	{
   		/// angle in degrees
		double degAngle;
	
		/// angle in radians
		double radAngle;
	
		/// tan of radAngle/2
		double dTan2;

		/// sin of radAngle
		double dSin;
	
		/// cos of radAngle
		double dCos;
	
		/// sin of -radAngle
		double dSinOpposite;
	
		/// cos of -radAngle
		double dCosOpposite;
	
		// Coordinate system
		double xc, yc;
		double XC, YC;
	};

	// computes rotation matrix
	void computeSinCosin(ipRotateMatrix2D *r, double d_angle);

	// rotation angle in degree
   	real degAngle;

	// opposite of rotation angle in degree
	real degAngleOpposite;

	// center of the rotation
	int xc, yc;

	// rotation matrices
	ipRotateMatrix2D rim;
	ipRotateMatrix2D rsim;

	// ipCore to shift the image
	ipShift *shift;

	// shift parameters
	int shiftx;
	int shifty;
	
	// internal variables
	int width1_;
	int height1_;
	real *pixmap1_;

	int width2_;
	int height2_;
	real *pixmap2_;
	
	int width3_;
	int height3_;
	real *pixmap3_;

	// BBx of rotated corners of the image
	sPoint2D bbx[5];	

	//
	void HorizSkew_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, int uRow, int iOffset, double dWeight);
	void VertSkew_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, int uCol, int iOffset, double dWeight);
	bool rotateimage0_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_);
	bool rotateimage90_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_);
	bool rotateimage180_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_);
	bool rotateimage270_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_);
	void rotateimage45_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, double dAngle);
	bool rotateimage_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_);

	//
	void HorizSkew_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, int uRow, int iOffset, double dWeight);
	void VertSkew_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, int uCol, int iOffset, double dWeight);
	bool rotateimage0_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_);
	bool rotateimage90_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_);
	bool rotateimage180_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_);
	bool rotateimage270_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_);
	void rotateimage45_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, double dAngle);
	bool rotateimage_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_);

	// background color
	rgbReal clrBack;

public:
	/** constructor
	
	    @param degAngle_ is the angle of rotation (in degree)
	    @param width_in_ is the #width# of the input image
	    @param height_in_ is the #height# of the input image
	    @param coding_in_ is the #coding# format of the input image
	*/
    	ipRotate(real degAngle_, int width_in_, int height_in_, const char *coding_);

	///
	virtual ~ipRotate();

	///
	virtual void process(Sequence *seq_in);

	//-----
	
	///
	virtual int getWidthOut(void);

	///
        virtual int getHeightOut(void);
	
	//-----
	
	/// set center of rotation
	void setCenter(int xc_, int yc_) { xc = xc_; yc = yc_; };

	/// rotate a point (integer)
	void rotate(int x, int y, int *X, int *Y);

	/// rotate a point (double)
	void rotate(double x, double y, double *X, double *Y);

	/// rotate a point in the opposite direction (integer)
	void unrotate(int x, int y, int *X, int *Y);

	/// rotate a point in the opposite direction (double)
	void unrotate(double x, double y, double *X, double *Y);
};

}

#endif
