#include "ipWeighedGaussian.h"
#include "ImageGray.h"

namespace Torch {

ipWeighedGaussian::ipWeighedGaussian(int size_, int width_, int height_, const char *coding_) : ipCore(coding_)
{
  size = size_;
  width = width_;
  height = height_;

  halfsize = (int)ceil(size/2);

  // variance = size/5
  sigma = (real)size/5;

  kernel = (real*) allocator->alloc(size*size*sizeof(real));
  w_kernel = (real*) allocator->alloc(size*size*sizeof(real));

  // build the original gaussian kernel
  for(int i = 0; i < size; i++) {
    
    real *kernel_line = kernel + (i*size);  // define a pointer on the kernel current line
    
    for (int j = 0; j < size; j++) {
      
      int x_value = (-halfsize) + i;
      int y_value = (-halfsize) + j;
      
      kernel_line[j] = exp(-(((real) (x_value * x_value + y_value * y_value)) / sigma));
    }
  }


  // allocate memory for output
  seq_out = new(allocator) Sequence(1, width*height);

  for(int i=0; i<seq_out->n_frames; i++)
     for(int j=0; j<seq_out->frame_size; j++)
       seq_out->frames[i][j] = 0.0; 
}


void ipWeighedGaussian::process(Sequence *seq_in)
{
 switch(coding_id)
    {
    case 0:
      {	
	real *src = seq_in->frames[0];
	
	
	// -------------------------------------------------------------------------------
	// CONVOLUTION
	// -------------------------------------------------------------------------------
	int k, kprime, l, lprime, index_k, index_l;
	int index_out = 0;
	real conv_sum = 0.0;

	for(k = 0; k < height; k++){
	  for(l = 0; l < width; l++){

	    // re-init weighed gaussian kernel (different weighting for each pixel)
	    for (int i=0; i<size*size; i++)
	      w_kernel[i]= kernel[i];
	    
	    real region_mean = 0;
	    int under = 0; 
	    int over = 0;
	    bool above;
	    conv_sum = 0.0;


	    // FIRST LOOP: compute region_mean 
	    for(kprime = -halfsize; kprime <= halfsize; kprime++){
	       
	      index_k = k - kprime;	
	      if(index_k < 0)
		index_k = abs(index_k) - 1;
	      if(index_k >= height)
		index_k = (2 * height) - index_k - 1; 
	       
	       
	      for(lprime = -halfsize; lprime <= halfsize; lprime++){
		 
		index_l = l-lprime;
		if(index_l < 0)
		  index_l = abs(index_l) - 1;
		if(index_l >= width)
		  index_l = (2 * width) - index_l - 1;	 

		real *image_line = src + (index_k*width); // define a pointer on the current image line
		
		region_mean += image_line[index_l];
	      }
	    }
	    region_mean /= (real)(size*size);

	    
	    
	    // SECOND LOOP: count number of pixels bigger/smaller than the mean
	    for(kprime = -halfsize; kprime <= halfsize; kprime++){
	       
	      index_k = k - kprime;	
	      if(index_k < 0)
		index_k = abs(index_k) - 1;
	      if(index_k >= height)
		index_k = (2 * height) - index_k - 1; 
	       
	       
	      for(lprime = -halfsize; lprime <= halfsize; lprime++){
		 
		index_l = l-lprime;
		if(index_l < 0)
		  index_l = abs(index_l) - 1;
		if(index_l >= width)
		  index_l = (2 * width) - index_l - 1;	 

		real *image_line = src + (index_k*width);
		 
		if(image_line[index_l] > region_mean)
		  over++;
		else
		  under++;
	      }
	    }
	    if (over > under)
	      above = true;
	    else
	      above = false; 
  
	    
	    
	    // THIRD LOOP: update filter weights	    
	    real norm_kernel = 0.0;
	    
	    for(kprime = -halfsize; kprime <= halfsize; kprime++){
	       
	      index_k = k - kprime;	
	      if(index_k < 0)
		index_k = abs(index_k) - 1;
	      if(index_k >= height)
		index_k = (2 * height) - index_k - 1; 
	       
	       
	      for(lprime = -halfsize; lprime <= halfsize; lprime++){
		 
		index_l = l-lprime;
		if(index_l < 0)
		  index_l = abs(index_l) - 1;
		if(index_l >= width)
		  index_l = (2 * width) - index_l - 1;	 
		 
		real *image_line = src + (index_k*width); 
		real *w_kernel_line = w_kernel + (kprime+halfsize)*size; // define a pointer on the current kernel line
		
	
	 	if (((image_line[index_l] > region_mean) && above == false) || ((image_line[index_l] < region_mean) && above == true)){
		  w_kernel_line[(lprime+halfsize)] = 0.0;		 
		}
		
		norm_kernel += w_kernel_line[(lprime+halfsize)];
	      }
	    }


	    // normalize s.t. sum(filter coeff) = 1
	    real weighed_sum = 0.0;
	    for (int i=0; i<size*size;i++)
	       weighed_sum += w_kernel[i];

	    for (int i=0; i<size*size;i++)
	      w_kernel[i] /= weighed_sum;



	    // FOURTH LOOP : convolve
	    for(kprime = -halfsize; kprime <= halfsize; kprime++){
	  
	      // mirror interpolation (if needed)
	      index_k = k - kprime;	
	      if(index_k < 0)
		index_k = abs(index_k) - 1;
	      if(index_k >= height)
		index_k = (2 * height) - index_k - 1; 
	      
	  
	      for(lprime = -halfsize; lprime <= halfsize; lprime++){
		
		// mirror interpolation (if needed)
		index_l = l-lprime;
		if(index_l < 0)
		  index_l = abs(index_l) - 1;
		if(index_l >= width)
		  index_l = (2 * width) - index_l - 1;
		
		real *image_line = src + (index_k*width);
		real *w_kernel_line = w_kernel + (kprime+halfsize)*size;
			
		conv_sum += w_kernel_line[(lprime+halfsize)]*image_line[index_l]; 
	      }
	    }	   
	    seq_out->frames[0][index_out]= conv_sum;
	    index_out++; 
	    conv_sum = 0;
	  } 
	}

	break;
      }
    case 1:
      message("ipWeighedGaussian::process() not implemented for RGB format");
      break;
    case 2:
      message("ipWeighedGaussian::process() not implemented for YUV format");
      break;
    case 3:
      message("ipWeighedGaussian::process() not implemented for float format");
      break;
    }
}

ipWeighedGaussian::~ipWeighedGaussian() 
{
}

}
