#include "multigrid.h"
#include "string.h"

//*************************************************************************************
//
// This file implements all the functions needed in mutigrid framework
// 
// -> restriction: fine-to-coarse projection (full weighting) 
// -> projection: coarse-to-fine projection (bilinear interpolation) 
// -> build differential operator
// -> compute diffusion coefficients for each pixel 
// -> customized matrix multiplication (takes advantage of the sparse structure)
// -> Relaxation: - Gauss-Seidel
//                - Jacobi 
//  
// 
// ---> author: Guillaume Heusch (heusch@idiap.ch) <---
//
//************************************************************************************


namespace Torch {


  void restrict(Vec *data, Vec *restricted, int width, int height) {

  
    
    int width_out = width / 2;
    int height_out = height / 2;
    
    int index_in = 0;
    int index_out = 0;
    
    
    for (index_out = 0; index_out < width_out*height_out; index_out++) {
      	
      int lin_out = (int)index_out/width_out;
      int col_out= index_out % width_out;

      // on the boundary: copy the corresponding pixel (injection)
      if ((lin_out == 0) || (lin_out == height_out-1) ||  (col_out == 0) || (col_out == width_out-1)){ 
	restricted->ptr[index_out] = data->ptr[index_in];
      }

      // full weighting scheme (interior points)
      else {
	
	restricted->ptr[index_out] = 0.25*data->ptr[index_in] + 0.125*(data->ptr[index_in-1] + data->ptr[index_in-width] + data->ptr[index_in+1] + data->ptr[index_in+width]) + 0.0625*(data->ptr[index_in-width-1] + data->ptr[index_in-width+1] + data->ptr[index_in+width-1] + data->ptr[index_in+width+1]); 

      }
      
      index_in += 2;
      
      if ((index_in % width) == 0)
	index_in += width;
    }
  }



  void project(Vec *data, Vec *projected, int width, int height) {
    
    int width_out = width*2;
    int height_out = height*2;

    int index_in = 0;

    int col_in, col_out, lin_in, lin_out;

    // process values that are directly copied
    for (index_in = 0; index_in < width*height; index_in++) {

      lin_in = (int)index_in/width;
      col_in = index_in % width;      
      col_out = 2*col_in;
      lin_out = 2*lin_in;

      projected->ptr[width_out*lin_out + col_out] = data->ptr[index_in];      
    }


    // process even lines and odd columns, interpolating horizontally
    for (int index_out = 0; index_out < width_out*height_out; index_out++) {

      bool is_on_boundary = (index_out < width_out) || (index_out >= (height_out-1)*width_out) ||  ((index_out % width_out) == 0) || ((index_out % width_out) == (width_out-1));
      
      // process pixel on the boundary
      if (is_on_boundary)
	projected->ptr[index_out] = 0.0;
      
      else {
									 
	// process pixel corresponding to even lines (0,2,4,...) and odd columns (1,3,5,..)
	int line = (int)index_out/width_out;
	int column = index_out % width_out; 
	
	if ((line % 2 == 0) && (column % 2 == 1))
	  projected->ptr[index_out] = 0.5*(projected->ptr[index_out+1] + projected->ptr[index_out-1]);
      }
    }

    
    // process odd lines, interpolating vertically
    for (int index_out = 0; index_out < width_out*height_out; index_out++) {

      bool is_on_boundary = (index_out < width_out) || (index_out >= (height_out-1)*width_out) ||  ((index_out % width_out) == 0) || ((index_out % width_out) == (width_out-1));
      
      // process pixel on the boundary
      if (is_on_boundary)
	projected->ptr[index_out] = 0.0;
    
      else {
	
	// process pixel corresponding to odd lines (1,3,5,...)
	int line = (int)index_out/width_out;
	
	if ((line % 2 == 1)) 
	  projected->ptr[index_out] = 0.5*(projected->ptr[index_out+width_out] + projected->ptr[index_out-width_out]);
      }
    }
  }


  void buildOperator(Mat *matrix, real *rho, real lambda, int type, Vec *image, int width, int height) {
    
    int mat_dim = width*height;
    matrix->zero();
    
    for (int i=0; i<mat_dim; i++) {
     
      // -- CHECK WHERE WE ARE -- i gives the pixel index
      bool up = true;
      bool down = true;
      bool right = true;
      bool left = true;
      
      // we are on the first image line
      if (i < width)
	up = false;
      
      // we are on the last image line    
      if (i >= ((height-1)*width)) 
	down = false;
      
      // we are on the first image column
      if ((i % width) == 0)
	left = false;
      
      // we are on the last image column
      if ((i % width) == (width-1))
	right = false;
     
 
      // -- FILL IN THE MATRIX ----------------------------------------------------
      matrix->ptr[i][i] = 1;
      
      // compute the diffusion coefficients associated to the current pixel (i)
      computeCoeff(rho, image, i, type, width, height);
      

      // upper diagonal
      if (right) {
	matrix->ptr[i][i+1] = -lambda*rho[RIGHT];
	matrix->ptr[i][i] += lambda*rho[RIGHT];
      }
      
      // lower diagonal
      if (left) {
	matrix->ptr[i][i-1] = -lambda*rho[LEFT];
	matrix->ptr[i][i] += lambda*rho[LEFT];
      }
      
      // upper fringe
      if (down) {
	matrix->ptr[i][i+width] = -lambda*rho[DOWN];
	matrix->ptr[i][i] += lambda*rho[DOWN];
      }
      
      // lower fringe
      if (up) {
	matrix->ptr[i][i-width] = -lambda*rho[UP];
	matrix->ptr[i][i] += lambda*rho[UP];
      }
    }
  }

  
  void computeCoeff(real *rho, Vec *image, int index, int type, int width, int height){
    
    bool up = true;
    bool down = true;
    bool left = true;
    bool right = true;

    // we are on the first image line
    if (index < width)
      up = false;
    
    // we are on the last image line    
    if (index >= ((height-1)*width)) 
      down = false;
    
    // we are on the first image column
    if ((index % width) == 0)
      left = false;
    
    // we are on the last image column
    if ((index % width) == (width-1))
      right = false;
    
    // initial contrast values set to zero, updated if we can compute it ...
    for (int i=0; i<5; i++)
      rho[i]= 0;


    // compute the five points stencil coefficients
    if (up) {
      
      switch(type) {
	
      case 0:
	rho[UP]=1;//*one_over_hy2;
	//print("rho up: %g\n", rho[UP]);
	break;
	
      case 1:
	peronaMalik(rho, image, index, width, height, "up");
	break;
	
      case 2:
	weber(rho, image, index, width, height, "up");
	break;	

      case 3:
	michelson(rho, image, index, width, height, "up");
	break;
      } 
    }


  if (down) {
    
    switch(type) {

      case 0:
	rho[DOWN]=1;//*one_over_hy2;
	break;
	
      case 1:
	peronaMalik(rho, image, index, width, height, "down");
	break;
	
      case 2:
	weber(rho, image, index, width, height, "down");
	break;	

      case 3:
	michelson(rho, image, index, width, height, "down");
	break; 

    } 
  }


  if (left) {
	
    switch(type) {
    case 0:
      rho[LEFT]=1;//*one_over_hx2;
	break;
	
      case 1:
	peronaMalik(rho, image, index, width, height, "left");
	break;
	
      case 2:
	weber(rho, image, index, width, height, "left");
	break;	

      case 3:
	michelson(rho, image, index, width, height, "left");
	break; 
   
    } 
  }


  if (right) {

    switch(type) {
 
      case 0:
	rho[RIGHT]=1;//*one_over_hx2;
	break;
	
      case 1:
	peronaMalik(rho, image, index, width, height, "right");
	break;
	
      case 2:
	weber(rho, image, index, width, height, "right");
	break;	

      case 3:
	michelson(rho, image, index, width, height, "right");
	break; 
    } 
  }
  
  rho[CENTER] = rho[UP] + rho[DOWN] + rho[LEFT] + rho[RIGHT];

  }

 



  void gaussSeidel(Vec *result, Vec *source, real *rho, real lambda, int type, int width, int height) {
    
    int dim = width*height;
    real up, down, left, right, center;

  
    // RED-BLACK GAUSS-SEIDEL

    for (int idx = 0; idx < dim; idx++) {

      // if we are  on the boundary
      if ((idx < width) || (idx >= (dim - width)) ||  ((idx % width) == 0) || ((idx % width) == (width-1))){ 
      
	// set the boundary values to zero
	result->ptr[idx] = 0.0;

      } 
    
      else {
      
	// if we are on an "even" pixel
	if (idx % 2 != 0) {
	
	  computeCoeff(rho, source, idx, type, width, height);

	  up = -lambda*rho[UP]*result->ptr[idx-width];
	  down = -lambda*rho[DOWN]*result->ptr[idx+width];
	  left = -lambda*rho[LEFT]*result->ptr[idx-1];
	  right = -lambda*rho[RIGHT]*result->ptr[idx+1];
	  center = 1 + lambda*rho[UP] + lambda*rho[DOWN] + lambda*rho[LEFT] + lambda*rho[RIGHT];
	
	  result->ptr[idx] = source->ptr[idx]/center - (up + down + left + right)/center;
	
	}
      }
    }
  
  
    for (int idx = 1; idx < dim; idx++) {

      // if we are not on the boundary
      if ((idx < width) || (idx >= (dim - width)) ||  ((idx % width) == 0) || ((idx % width) == (width-1))) { 
      
	// set the boundary values to zero
	result->ptr[idx] = 0.0; 
      } 
    
      else {
    
	// if we are on an "odd" pixel
	if (idx % 2 == 0) {

	  computeCoeff(rho, source, idx, type, width, height);

	  up = -lambda*rho[UP]*result->ptr[idx-width];
	  down = -lambda*rho[DOWN]*result->ptr[idx+width];
	  left = -lambda*rho[LEFT]*result->ptr[idx-1];
	  right = -lambda*rho[RIGHT]*result->ptr[idx+1];
	  center = 1 + lambda*rho[UP] + lambda*rho[DOWN] + lambda*rho[LEFT] + lambda*rho[RIGHT];

	  result->ptr[idx] = source->ptr[idx]/center - (up + down + left + right)/center;
	}
      }
    }
  
  }
    
  

  void myMultiply(Vec *data, Vec *result, real *rho, real lambda, int type, int width, int height) {
    
    for (int i=0; i<width*height; i++) {

      // BEWARE --- first check were we are
      bool up = true;
      bool down = true;
      bool left = true;
      bool right = true;
    
      real coeff_up = 0.0;
      real coeff_down = 0.0;
      real coeff_left = 0.0;
      real coeff_right = 0.0;
      real coeff_center = 1.0;
    
      // we are on the first image line
      if (i < width)
	up = false;

      // we are on the last image line    
      if (i >= ((height-1)*width)) 
	down = false;
    
      // we are on the first image column
      if ((i % width) == 0)
	left = false;
    
      // we are on the last image column
      if ((i % width) == (width-1))
	right = false;

      bool is_on_boundary = (!up || !down || !left || !right);

      if (is_on_boundary)
	result->ptr[i] = 0.0;

      else {
    
	computeCoeff(rho, data, i, type, width, height);
	
	if (up){
	  coeff_up = -lambda*rho[UP]*data->ptr[i-width];
	  coeff_center += lambda*rho[UP];
	}
	if (down) {
	coeff_down = -lambda*rho[DOWN]*data->ptr[i+width];
	coeff_center += lambda*rho[DOWN];
	}
	if (left) {
	coeff_left = -lambda*rho[LEFT]*data->ptr[i-1];
	coeff_center += lambda*rho[LEFT];
	}
	if (right) {
	  coeff_right = -lambda*rho[RIGHT]*data->ptr[i+1];
	coeff_center += lambda*rho[RIGHT];
	}
	
	result->ptr[i] = coeff_center*data->ptr[i] + coeff_up + coeff_down + coeff_left + coeff_right; 
      }
    }
  }


  void jacobi(Vec *result, Vec *source, real *rho, real lambda, int type, int width, int height) { 
    
    int dim = width*height;
    real up, down, left, right, center;
    Vec* old = new Vec(width*height); 

    old = result;
 
    real weight = 1; // weight used in damped Jacobi (set to 1 = pure Jacobi)

    for (int idx = 0; idx < dim; idx++) {
      
      // if we are  on the boundary
      if ((idx < width) || (idx >= (dim - width)) ||  ((idx % width) == 0) || ((idx % width) == (width-1))){ 
	
	// set the boundary values to zero
	result->ptr[idx] = 0.0;
	
      } 
    
      else {

	computeCoeff(rho, source, idx, type, width, height);

	up = -lambda*rho[UP]*old->ptr[idx-width];
	down = -lambda*rho[DOWN]*old->ptr[idx+width];
	left = -lambda*rho[LEFT]*old->ptr[idx-1];
	right = -lambda*rho[RIGHT]*old->ptr[idx+1];
	center = 1 + lambda*rho[UP] + lambda*rho[DOWN] + lambda*rho[LEFT] + lambda*rho[RIGHT];
	
	result->ptr[idx] = (1-weight)*old->ptr[idx] + weight*(source->ptr[idx]/center - (up + down + left + right)/center);
      } 
    }

  }



  void weber(real *rho, Vec *image, int idx, int width, int height, char *position) {

    if (strcmp(position, "up") == 0) {

      real center = image->ptr[idx];
      real up = image->ptr[idx-width];
      real diff = fabs(center - up);
      real min = 0.0;

      if (up > center)
	min = center; 
      else 
	min = up;

      if (IS_NEAR(diff, 0, 1))
	rho[UP] = min;
      else 
	rho[UP] = min/diff;
    }

    if (strcmp(position, "down") == 0) {

      real center = image->ptr[idx];
      real down = image->ptr[idx+width];
      real diff = fabs(center - down);
      real min = 0.0;

      if (down > center)
	min = center; 
      else 
	min = down;

      if (IS_NEAR(diff, 0, 1))
	rho[DOWN] = min;
      else 
	rho[DOWN] = min/diff;
    }

    if (strcmp(position, "left") == 0) {

      real center = image->ptr[idx];
      real left = image->ptr[idx-1];
      real diff = fabs(center - left);
      real min = 0.0;

      if (left > center)
	min = center; 
      else 
	min = left;

      if (IS_NEAR(diff, 0, 1))
	rho[LEFT] = min;
      else 
	rho[LEFT] = min/diff;
    }

    if (strcmp(position, "right") == 0) {

      real center = image->ptr[idx];
      real right = image->ptr[idx+1];
      real diff = fabs(center - right);
      real min = 0.0;

      if (right > center)
	min = center; 
      else 
	min = right;

      if (IS_NEAR(diff, 0, 1))
	rho[RIGHT] = min;
      else 
	rho[RIGHT] = min/diff;
    }
  }


  void michelson(real *rho, Vec *image, int idx, int width, int height, char *position) {

    if (strcmp(position, "up") == 0) {

      real center = image->ptr[idx];
      real up = image->ptr[idx-width];
      real diff = fabs(center - up);

      if (IS_NEAR(diff, 0, 1))
	rho[UP] = (center+up);
      else 
	rho[UP] = (center+up)/diff;
    
    }

    if (strcmp(position, "down") == 0) {

      real center = image->ptr[idx];
      real down = image->ptr[idx+width];
      real diff = fabs(center - down);
      
      if (IS_NEAR(diff, 0, 1))
	rho[DOWN] = (center+down);
      else 
	rho[DOWN] = (center+down)/diff;
     
    }

    if (strcmp(position, "left") == 0) {

      real center = image->ptr[idx];
      real left = image->ptr[idx-1];
      real diff = fabs(center - left);
     
      if (IS_NEAR(diff, 0, 1))
	rho[LEFT] = (center+left);
      else 
	rho[LEFT] = (center+left)/diff;
    }
    
    if (strcmp(position, "right") == 0) {
      
      real center = image->ptr[idx];
      real right = image->ptr[idx+1];
      real diff = fabs(center - right);
     
      if (IS_NEAR(diff, 0, 1))
	rho[RIGHT] = (center+right);
      else 
	rho[RIGHT] = (center+right)/diff;
    }
  } 
  


  void peronaMalik(real *rho, Vec *image, int idx, int width, int height, char *position) {
    
    real tmp1 = 0.0;
    real tmp2 = 0.0;
    real tmp3 = 0.0;
    real tmp4 = 0.0;
    
    int k=30;
    
    if (strcmp(position, "up") == 0) {
      
      tmp1 = (image->ptr[idx] - image->ptr[idx-width])/k;
      rho[UP] = exp(-(tmp1*tmp1));
    }
    
    if (strcmp(position, "down") == 0) {
      
      tmp2 = (image->ptr[idx] - image->ptr[idx+width])/k;
      rho[DOWN] = exp(-(tmp2*tmp2));
    }

    if (strcmp(position, "left") == 0) {
      
      tmp3 = (image->ptr[idx] - image->ptr[idx-1])/k;
      rho[LEFT] = exp(-(tmp3*tmp3));
    }
    
    if (strcmp(position, "right") == 0) {

      tmp4 = image->ptr[idx] - image->ptr[idx+1];
      rho[RIGHT] = exp(-(tmp4*tmp4));
    }       
 }

}
