#ifndef MULTIGRID_INC
#define MULTIGRID_INC

#include "Vec.h"
#include "Mat.h"

// used for the coefficients
#define CENTER 0
#define LEFT 1
#define RIGHT 2
#define UP 3
#define DOWN 4

#define IS_NEAR(var, value, delta) ((var >= (value - delta)) && (var <= (value + delta)))

namespace Torch {

/** @name Useful functions for the multigrid framework
 
    @author Guillaume Heusch (heusch@idiap.ch)
    \Date
    @version 2.0
    @since 2.0
*/
//@{

  /** intergrid transfer function: fine-to-coarse projection (full weighting)
    
  @param data a vector defined on a fine grid 
  @param restricted a vector on the coarse grid (result)
  @param width the width of the fine grid
  @param height the height of the fine grid 
  */
  void restrict(Vec *data, Vec *restricted, int width, int height);
  
  /** intergrid transfer function: coarse-to-fine projection (bilinear interpolation)
      
  @param data a vector defined on a coarse grid 
  @param restricted a vector on the fine grid (result)
  @param width the width of the coarse grid
  @param height the height of the coarse grid 
  */   
  void project(Vec *data, Vec *projected, int width, int height);
  
  /** build the differential operator (i.e. the sparse matrix induced by the discretization of the PDE

  @param matrix  the differential operator (result)
  @param rho the diffusion coefficients 
  @param lambda the relative importance of the smoothness constraint
  @param type the the kind of diffusion performed (isotropic, anisotropic)
  @param image the data
  @param width data's width 
  @param height data's height 
  */
  void buildOperator(Mat *matrix, real *rho, real lambda, int type, Vec *image, int width, int height);
  
  /** compute the diffusion coefficients for each point of the data
      
  @param index the index of the processed point
  */
  void computeCoeff(real *rho, Vec *image, int index, int type, int width, int height);
  
  /** perform Gauss-Seidel relaxation on Ax=b

  @param source is the right hand side term of the above equation (b)
  @param rho the diffusion coefficients 
  @param lambda the relative importance of the smoothness constraint
  @param type the the kind of diffusion performed (isotropic, anisotropic)
  */
  void gaussSeidel(Vec *result, Vec *source, real *rho, real lambda, int type, int width, int height);

  /// perform Jacobi relaxation (see Gauss-Seidel for parameters)
  void jacobi(Vec *result, Vec *source, real *rho, real lambda, int type, int width, int height); 
  
  /// custom matrix multiplication (takes advantage of the sparse structure of the differential operator)
  void myMultiply(Vec *data, Vec *result, real *rho, real lambda, int type, int width, int height);

  /** compute, for each point, the diffusion coefficients (weber contrast)
  
  @param rho array with the coefficients for a point (result)
  @param image the data
  @param idx index of the point 
  @param position direction in which the coefficient is computed (up, down, left, right)
  */
  void weber(real *rho, Vec *image, int idx, int width, int height, char *position); 

  /// compute the diffusion coefficients (micheslon contrast) (see weber for parameters)
  void michelson(real *rho, Vec *image, int idx, int width, int height, char *position); 

  /// compute the diffusion coefficients (perona-malik exponential) (see weber for parameters)
  void peronaMalik(real *rho, Vec *image, int idx, int width, int height, char *position);

  //@}
}
#endif
