#include "DiskXFile.h"

#include "PreProcessing.h"

#include "Linear.h"
#include "Tanh.h"
#include "Sigmoid.h"
#include "SoftMax.h"
#include "LogSoftMax.h"
#include "Exp.h"
#include "SoftPlus.h"

#include "MyMLP.h"

namespace Torch {

struct sMyMLP_layerformat_
{
	char *layerformat;
	GradientMachine* (*layercreate)(Allocator *allocator, int n_inputs_, int n_outputs_);
};

GradientMachine* linear_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* tanh_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* sigmoid_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* softmax_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* logsoftmax_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* exp_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);
GradientMachine* softplus_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_);


static int n_layerformat = 7;
sMyMLP_layerformat_ MyMLP_layerformat[]={ 	{"linear", &linear_layer_create}, // Linear should be the first one
                                 		{"tanh", &tanh_layer_create},
                                 		{"sigmoid", &sigmoid_layer_create},
                                 		{"softmax", &softmax_layer_create},
                                 		{"log-softmax", &logsoftmax_layer_create},
                                 		{"exp", &exp_layer_create},
						{"softplus", &softplus_layer_create}};

int find_layerformat(const char *layerformat_)
{
 	for(int i = 0 ; i < n_layerformat ; i++)
        {
                if(strcmp(layerformat_, MyMLP_layerformat[i].layerformat) == 0)
                {
                	message("[OK] %s", MyMLP_layerformat[i].layerformat);
                         
                        return i;
                }
        }
  
	return -1;  
}


MyMLP::MyMLP(int n_layers_, int n_inputs_, ...)
{
   	verbose = false;

	if(n_layers_ < 2) error("MyMLP::MyMLP() n_layers should be >= 2", n_layers_);
	
  	n_layers = n_layers_;
  	layers = (GradientMachine **)allocator->alloc(sizeof(GradientMachine *)*n_layers);
  	is_linear = (bool *)allocator->alloc(sizeof(bool)*n_layers);
  	for(int i = 0; i < n_layers; i++) is_linear[i] = false;

	layerformat = (int *)allocator->alloc(sizeof(int)*n_layers);
	nhu = (int *)allocator->alloc(sizeof(int)*n_layers);
	
  	va_list args;
  	va_start(args, n_inputs_);  
  	for(int i = 0; i < n_layers; i++)
  	{
    		char *layer_type = va_arg(args, char *);
    		int n_outputs_ = va_arg(args, int);
    		bool is_valid = false;

		int lf = find_layerformat(layer_type);

		if(lf != -1)
		{
			if(lf != 0) // Not linear 
			{
  				//layers = (GradientMachine **)allocator->realloc(layers, sizeof(GradientMachine *)*(n_layers+1));

			   	// Create a linear layer first
		   		layers[i] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
				is_linear[i] = true;
    				this->addFCL(layers[i]);
				
		   		//layers[i] = MyMLP_layerformat[lf].layercreate(allocator, n_inputs_, n_outputs_);
		   		GradientMachine *machine_ = MyMLP_layerformat[lf].layercreate(allocator, n_inputs_, n_outputs_);
				this->addFCL(machine_);
			}
			else
			{
		   		layers[i] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
				is_linear[i] = true;
    				this->addFCL(layers[i]);
			}
			
      			is_valid = true;
			layerformat[i] = lf;
			nhu[i] = n_outputs_;
		}

    		if(!is_valid)
      			error("MLP: unknow layer type <%s>", layer_type);
    
    		//this->addFCL(layers[i]);
    		n_inputs_ = n_outputs_;
  	}

	//
  	build();

	//
	dump_params_init();
	
  	va_end(args);
}
    	
MyMLP::MyMLP(int n_inputs_, char *layer_, int n_outputs_)
{
   	verbose = false;
  	
	n_layers = 1;
  	layers = (GradientMachine **)allocator->alloc(sizeof(GradientMachine *)*n_layers);
	layers[0] = NULL;
  	is_linear = (bool *)allocator->alloc(sizeof(bool)*n_layers);
  	is_linear[0] = false;
	
	layerformat = (int *)allocator->alloc(sizeof(int)*n_layers);
	nhu = (int *)allocator->alloc(sizeof(int)*n_layers);

	nhu[0] = 0;
	
	int lf = find_layerformat(layer_);

	if(lf != -1)
	{
		if(lf != 0) // Not linear 
		{
			// Create a linear layer first
		   	layers[0] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
			is_linear[0] = true;
    			this->addFCL(layers[0]);
				
		   	GradientMachine *machine_ = MyMLP_layerformat[lf].layercreate(allocator, n_inputs_, n_outputs_);
    			this->addFCL(machine_);
		}
		else
		{	
			// Only a linear
		   	layers[0] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
			is_linear[0] = true;
    			this->addFCL(layers[0]);
		}
			
		layerformat[0] = lf;
	}
      	else error("MLP: unknow layer type <%s>", layer_);
   
	//n_inputs = n_inputs_;
	//n_outputs = n_outputs_;

	//
  	build();

	//
	dump_params_init();
}

MyMLP::MyMLP(const char *filename_)
{
   	verbose = false;

   	layerformat = NULL;
	nhu = NULL;
	layers = NULL;
	is_linear = NULL;
	params_init = NULL;

	load(filename_);
}

MyMLP::MyMLP()
{
   	verbose = false;

   	layerformat = NULL;
	nhu = NULL;
	layers = NULL;
	is_linear = NULL;

	params_init = NULL;
}

void MyMLP::setWeightDecay(real weight_decay)
{
  	for(int i = 0; i < n_layers; i++)
    		if(is_linear[i]) layers[i]->setROption("weight decay", weight_decay);
}

void MyMLP::info()
{
	message("Number of parameters: %d", params->n_params);
  	message("n_inputs = %d", n_inputs);
  	message("n_layer = %d", n_layers);
  	for(int i = 0; i < n_layers; i++)
		message("   layer [%d] = %dx%s", i, nhu[i], MyMLP_layerformat[layerformat[i]].layerformat);
  	message("n_outputs = %d", n_outputs);
}
	
void MyMLP::dump_params_init()
{
   	message("MyMLP::dump_params_init()");
	
   	int n_ = 0;

	// determine the total number of object Parameters
	//print("original parameters = {");
	if(params)
		for(int i = 0 ; i < params->n_data; i++)
			for(int j = 0; j < params->size[i]; j++)
			{
				//print("%g ", params->data[i][j]);
			
	  			n_++;
			}
	//print("} [%d, %d]\n", n_, params->n_params);
	
	if(n_ != params->n_params) error("MyMLP::dump_params_init() incorrect number of parameters");
	   
	// creates an array #params_init# of Parameters
	params_init = (real *) allocator->alloc(params->n_params * sizeof(real));

	// copy the current parameters to #params_init#
	n_ = 0;
	if(params)
		for(int i = 0 ; i < params->n_data; i++)
			for(int j = 0; j < params->size[i]; j++)
				params_init[n_++] = params->data[i][j];
}

void MyMLP::reset()
{
   	//message("MyMLP::reset()");
   	//ConnectedMachine::reset();

	real *params_init_ = params_init;

	//print("initial parameters = {");
	if(params)
		for(int i = 0 ; i < params->n_data; i++)
			for(int j = 0; j < params->size[i]; j++)
			{
				params->data[i][j] = *params_init_++;
				//print("%g ", params->data[i][j]);
			}
	//print("}\n");
}

void MyMLP::load(const char *filename_, PreProcessing *preprocess_norm)
{
  	if(verbose) message("MyMLP::load(%s) ...", filename_);
	
	DiskXFile file(filename_, "r");

  	loadXFile(&file);

	if(preprocess_norm != NULL)
	{
		//
		// Load normalisation
  		if(verbose) print("  loading Normalisation ...\n");
        	preprocess_norm->loadXFile(&file);
	}
}

void MyMLP::save(const char *filename_, PreProcessing *preprocess_norm)
{
  	if(verbose) message("MyMLP::save(%s) ...", filename_);
	
	DiskXFile file(filename_, "w");

  	saveXFile(&file);

	if(preprocess_norm != NULL)
	{
		//
		// Save normalisation
  		if(verbose) print("  saving Normalisation ...\n");
        	preprocess_norm->saveXFile(&file);
	}
}

void MyMLP::loadXFile(XFile *file)
{
	int lf;
	   
	//
	// Load first MLP architecture
  	if(verbose) print("  loading Architecture ...\n");
	file->taggedRead(&n_inputs, sizeof(int), 1, "N_INPUTS");
	file->taggedRead(&n_layers, sizeof(int), 1, "N_LAYERS");

	//
	layers = (GradientMachine **)allocator->alloc(sizeof(GradientMachine *)*n_layers);
  	is_linear = (bool *)allocator->alloc(sizeof(bool)*n_layers);
  	for(int i = 0; i < n_layers; i++) is_linear[i] = false;
	
	//
	if(layerformat != NULL) allocator->free(layerformat);
	if(nhu != NULL) allocator->free(nhu);
  	
	layerformat = (int *)allocator->alloc(sizeof(int)*n_layers);
	nhu = (int *)allocator->alloc(sizeof(int)*n_layers);

	//
	if(n_layers == 1)
	{
	   	int n_outputs_;

		file->taggedRead(&lf, sizeof(int), 1, "LAYERFORMAT");
		file->taggedRead(&n_outputs_, sizeof(int), 1, "NHU");

		if(n_outputs_ != 0) warning("MyMLP::loadXFile() NHU=0 was expected here");
		
		file->taggedRead(&n_outputs_, sizeof(int), 1, "N_OUTPUTS");
	
		if(lf != -1)
		{
			if(lf != 0) // Not linear 
			{
			   	// Create a linear layer first
		   		layers[0] = MyMLP_layerformat[0].layercreate(allocator, n_inputs, n_outputs_);
				is_linear[0] = true;
    				this->addFCL(layers[0]);
				
		   		GradientMachine *machine_ = MyMLP_layerformat[lf].layercreate(allocator, n_inputs, n_outputs_);
    				this->addFCL(machine_);
			}
			else
			{
		   		layers[0] = MyMLP_layerformat[0].layercreate(allocator, n_inputs, n_outputs_);
				is_linear[0] = true;
    				this->addFCL(layers[0]);
			}
			
			layerformat[0] = lf;
			nhu[0] = 0;
		}
	}
	else
	{
		int n_inputs_, n_outputs_;

		n_inputs_ = n_inputs;

  		for(int i = 0; i < n_layers; i++)
		{
			file->taggedRead(&lf, sizeof(int), 1, "LAYERFORMAT");
			file->taggedRead(&n_outputs_, sizeof(int), 1, "NHU");

    			bool is_valid = false;

			if(lf != -1)
			{
				if(lf != 0) // Not linear 
				{
  					//layers = (GradientMachine **)allocator->realloc(layers, sizeof(GradientMachine *)*(n_layers+1));

			   		// Create a linear layer first
		   			layers[i] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
					is_linear[i] = true;
    					this->addFCL(layers[i]);
				
		   			//layers[i] = MyMLP_layerformat[lf].layercreate(allocator, n_inputs_, n_outputs_);
		   			GradientMachine *machine_ = MyMLP_layerformat[lf].layercreate(allocator, n_inputs_, n_outputs_);
    					this->addFCL(machine_);
				}
				else
				{
		   			layers[i] = MyMLP_layerformat[0].layercreate(allocator, n_inputs_, n_outputs_);
					is_linear[i] = true;
    					this->addFCL(layers[i]);
				}
			
      				is_valid = true;
				layerformat[i] = lf;
				nhu[i] = n_outputs_;
			}

    			if(!is_valid)
      				error("MLP: unknow layer type <%s>", MyMLP_layerformat[lf].layerformat);
    
    			//this->addFCL(layers[i]);
    			n_inputs_ = n_outputs_;
  		}
	
		file->taggedRead(&n_outputs, sizeof(int), 1, "N_OUTPUTS");
	}

	build();

	//
	// Load parameters
  	if(verbose) print("  loading Parameters ...\n");
	ConnectedMachine::loadXFile(file);
}

void MyMLP::saveXFile(XFile *file)
{
	//
	// Save first MLP architecture
  	if(verbose) print("  saving Architecture ...\n");
	
	file->taggedWrite(&n_inputs, sizeof(int), 1, "N_INPUTS");
	file->taggedWrite(&n_layers, sizeof(int), 1, "N_LAYERS");
  	for(int i = 0; i < n_layers; i++)
	{
		file->taggedWrite(&layerformat[i], sizeof(int), 1, "LAYERFORMAT");
		file->taggedWrite(&nhu[i], sizeof(int), 1, "NHU");
	}
	file->taggedWrite(&n_outputs, sizeof(int), 1, "N_OUTPUTS");
	
	//
	// Save parameters
  	if(verbose) print("  saving Parameters ...\n");
	ConnectedMachine::saveXFile(file);
}


MyMLP::~MyMLP()
{
}

GradientMachine *linear_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;

//	print("+ Linear\n");
	
	ptr = new(allocator) Linear(n_inputs_, n_outputs_);

	return ptr;
}
   
GradientMachine* tanh_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ Tanh\n");
	
	ptr = new(allocator) Tanh(n_outputs_);

	return ptr;
}

GradientMachine* sigmoid_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ Sigmoid\n");
	
	ptr = new(allocator) Sigmoid(n_outputs_);

	return ptr;
}

GradientMachine* softmax_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ SoftMax\n");
	
	ptr = new(allocator) SoftMax(n_outputs_);

	return ptr;
}

GradientMachine* logsoftmax_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ LogSoftMax\n");
	
	ptr = new(allocator) LogSoftMax(n_outputs_);

	return ptr;
}

GradientMachine* exp_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ Exp\n");
	
	ptr = new(allocator) Exp(n_outputs_);

	return ptr;
}

GradientMachine* softplus_layer_create(Allocator *allocator, int n_inputs_, int n_outputs_)
{
   	GradientMachine *ptr = NULL;
	
//	print("+ SoftPlus\n");
	
	ptr = new(allocator) SoftPlus(n_outputs_);

	return ptr;
}

}
