#ifndef PATTERN_DETECT_INC
#define PATTERN_DETECT_INC

#include "WindowScan.h"
#include "ipIntegralImage.h"

namespace Torch {
   
/** This class is designed to scan an image at several scales and locations

    \begin{verbatim}
    
	       width_image
	+----------------------+
	|   Wmin               |
	|  +-+                 |
	|  | |                 |
	|  | |         Wmax    |
	|  +-+      +--+       | height_image
	|           |  |       |
	|           |  |       |
	|           |  |       |
	|           +--+       |
	+----------------------+

    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @author Yann Rodriguez (rodrig@idiap.ch)
    \Date
    @version 2.0
    @since 2.0
*/
class PatternDetect : public Object 
{
protected:
	//
	bool verbose;
	
	// size of the image to scan and ratio H/W
	int width_image;
        int height_image;
        real ratioImage;

	// size of the training pattern model and ratio H/W
	int patternWmin;
        int patternHmin;
        real ratioPattern;
	
	// scale and step factors
	real scale_factor;
	real stepx_factor, stepy_factor;

	// scanning window min/max size (determines number of scale levels)
	int Wmin;
	int Wmax;

	// scale levels
	int n_levels;
	WindowScan **levels;
	
	// Integral Images
	ipIntegralImage *ip_ii;
	
	// ipSubwindow
	ipSubWindow *ip_sw;
	
	// mean-stdv filter
	real minMean, maxMean, minStdv, maxStdv;

	// pruning filter
	real *pruning_ii;
        real pruning_min_threshold, pruning_max_threshold;

	// check if init done
	bool init_done;

public:
	/// the pattern collector
	PatternCollector *patterncollector;
	
	/// number of output patterns (detected and merged)
	int n_patterns;

	/// output patterns (detected and merged)
	Pattern **patterns;

	/// total number of hypothesis
	int n_subwindows;
	
	/// number of hypothesis tested
	int n_subwindows_processed;

	/// number of hypothesis rejected by prunning
	int n_pruning_reject;

	/// number of hypothesis rejected by mean
	int n_mean_reject;

	/// number of hypothesis rejected by stdv
	int n_stdv_reject;
	
	//-----
	
	/** detect patterns in an image using an ipSubwindow
	
	    @param width_image_ is the #width# of the image
	    @param height_image_ is the #height# of the image
	    @param patternWmin_ is the minimum #width# of the pattern
	    @param patternHmin_ is the minimum #height# of the pattern
	    @param Wmin_ is the minimum #width# of the subwindow to scan
	    @param Wmax_ is the maximum #width# of the subwindow to scan
	    @param stepx_ is the x step for scanning
	    @param stepy_ is the y step for scanning
	    @param scale_factor_ is the scale factor
	    @param minMean_ is the minimum mean value of the pixels in a subwindow
	    @param maxMean_ is the maximum mean value of the pixels in a subwindow
	    @param minStdv_ is the minimum stdv value of the pixels in a subwindow
	    @param maxStdv_ is the maximum stdv value of the pixels in a subwindow
	    @param pruning_ii_ is the integral image of a binary pruning mask
	    @param pruning_min_threshold_ is the min number of 1 in #pruning_ii_#
	    @param pruning_max_threshold_ is the max number of 1 in #pruning_ii_#
	    @param max_pattern is the maximum number of patterns that can be stored
	*/
	PatternDetect(	int width_image_, int height_image_, 
			int patternWmin_, int patternHmin_, 
			int Wmin_, int Wmax_, real stepx_, real stepy_,
			real scale_factor_,
			real minMean_, real maxMean_, real minStdv_, real maxStdv_,
			real *pruning_ii_,
			real pruning_min_threshold_,
			real pruning_max_threshold_,
			int max_pattern=10000,
			bool verbose=false);

	///
	virtual void init(void);

	///
	virtual void process(Sequence *seq_in);

	///
	virtual void stat(void);

	///
	int getWmin() { return Wmin; };

	///
	int getWmax() { return Wmax; };

	///
	virtual void mergepatterns(void) = 0;

	///
	virtual ~PatternDetect();
};

}

#endif
