#ifndef WINDOW_SCAN_INC
#define WINDOW_SCAN_INC

#include "PatternCollector.h"
#include "ipSubWindow.h"

namespace Torch {
   
/** This class is designed to scan an image with a fixed subwindow size
 
    \begin{verbatim}

	          width_image
	        (real_width_image)
	+-------------------------+
	|                         |
	|    width_pattern        |
	|  +---+                  |
	|  |   |                  | height_image
	|  |   | height_pattern   | (real_height_image)
	|  |   |                  |
	|  +---+                  |
	|                         |
	|                         |
	|   +---+       +---+     |
	|   |   |       |   |     |
	|   |   |  ...  |   |     |
	|   |   |       |   |     |
	|   +---+       +---+     |
	|                         |
	|                         |
	+-------------------------+
 
    \end{verbatim}

    @author Sebastien Marcel (marcel@idiap.ch)
    @author Yann Rodriguez (rodrig@idiap.ch)
    \Date
    @version 2.0
    @since 1.0
*/
class WindowScan : public Object 
{
protected:
   	//
	bool verbose;
	
	// size of the image to scan
	int width_image;
	int height_image;

	// real size of the image to scan (if subsampled)
	int real_width_image;
	int real_height_image;

	// size of the pattern to scan over the image
	int width_pattern;
	int height_pattern;

	// real size of the pattern to scan over the image (if subsampled)
	int real_width_pattern;
	int real_height_pattern;

	// x and y steps for window translation
	int stepx;
	int stepy;

	// machine to process the subwindows
	ipSubWindow *ip_sw;
	
	// pattern collector to collect patterns
	PatternCollector *patterncollector;

	// integral images, mean and stdv preprocessing
	bool use_mean_stdv;
	real *integralimage;
        real *integralimage2;
        real minMean, maxMean;
        real minStdv, maxStdv;
        real one_by_size_pattern;

	// integral image of a binary pruning map
	bool use_pruning;
	real *pruning_ii;
	real pruning_min_threshold, pruning_max_threshold;
	
public:
	/// total number of hypothesis
	int n_subwindows;
	
	/// number of hypothesis tested
	int n_subwindows_processed;

	/// number of hypothesis rejected by prunning
	int n_pruning_reject;

	/// number of hypothesis rejected by mean
	int n_mean_reject;

	/// number of hypothesis rejected by stdv
	int n_stdv_reject;

	//---
	
	/** scan an image at a fixed window size
	
	    The image to process could be a subsampled image. 
	    The original image is referred here as real image.
	    
	    @param width_image_ is the #width# of the image to process
	    @param height_image_ is the #height# of the image to process
	    @param real_width_image_ is the real #width# of the image
	    @param real_height_image_ is the real #height# of the image
	    @param width_pattern_ is the #width# of the subwindow to scan
	    @param height_pattern_ is the #height# of the subwindow to scan
	    @param real_width_pattern_ is the real #width# of the subwindow in the real image
	    @param real_height_pattern_ is the real #height# of the subwindow in the real image
	    @param stepx_factor_ is the #x# step for scanning
	    @param stepy_factor_ is the #y# step for scanning
	    @param ip_sw_ is the ipSubWindow to apply to each subwindow (that pass prunning)
	    @param patterncollector_ is the PatternCollector to use to collect patterns
	*/
	WindowScan(	int width_image_, int height_image_, 
			int real_width_image_, int real_height_image_, 
			int width_pattern_, int height_pattern_, 
			int real_width_pattern_, int real_height_pattern_, 
			real stepx_factor_, real stepy_factor_, 
			ipSubWindow *ip_sw_,
			PatternCollector *patterncollector_,
			bool verbose = false);

	///
	virtual void init(void) = 0;

	///
	virtual void process(Sequence *seq_in) = 0;

	/// perform pruning of subwindows according to the mean and the stdv of pixels
	void setMeanStdv(real minMean_, real maxMean_, real minStdv_, real maxStdv_, real *integralimage_, real *integralimage2_);

	/// perform prunning of subwindows according to the integral image of a binary mask
	void setPruningII(real pruning_min_threshold, real pruning_max_threshold, real *pruning_ii);

	///
	virtual ~WindowScan();
};

}

#endif
